# -*- coding: utf-8 -*-


def split_load(list_submodules, chunks):
    list_submodules_list = list()
    num_jobs = 0
    while num_jobs < chunks:
        list_submodules_list.append(list())
        num_jobs += 1

    i = 0
    for submodule in list_submodules:
        list_submodules_list[i % num_jobs].append(submodule)
        i += 1
    return list_submodules_list



# -*- coding: utf-8 -*-

from threading import Lock
import copy


class Counter(object):
    def __init__(self, num_elements=0):
        self.__value = 0
        self.__num_elements = num_elements
        self.__lock = Lock()

    def get_value(self):
        self.__lock.acquire()
        val = copy.copy(self.__value)
        self.__lock.release()
        return val

    def set_value(self, val):
        self.__lock.acquire()
        self.__value = val
        self.__lock.release()

    def reset(self):
        self.__lock.acquire()
        self.__value = 0
        self.__lock.release()

    def increment_value(self, increment=1):
        self.__lock.acquire()
        self.__value = self.__value + increment
        val = copy.copy(self.__value)
        self.__lock.release()
        if self.__num_elements == 0:
            return val
        else:
            return "[" + str(val) + "/" + str(self.__num_elements) + "]"

    def decrement_value(self, decrement=1):
        self.__lock.acquire()
        self.__value = self.__value + decrement
        val = copy.copy(self.__value)
        self.__lock.release()
        if self.__num_elements == 0:
            return val
        else:
            return "[" + str(val) + "/" + str(self.__num_elements) + "]"



# -*- coding: utf-8 -*-

from parallelforeachsubmodule.metadata import Metadata
from parallelforeachsubmodule.process import PFSProcess
from parallelforeachsubmodule.status import Counter
from parallelforeachsubmodule.chunker import split_load
from parallelforeachsubmodule.scheduler import Scheduler
import argparse
import threading
import os
import re
import time
import multiprocessing


def worker(submodule_list, path, command, counter, output_filter="", cmd_func=None):
    if isinstance(submodule_list, Scheduler):
        while not submodule_list.empty():
            PFSProcess(submodule_list.get(), path, command, counter, output_filter, cmd_func).run()
    else:
        for submodule in submodule_list:
            PFSProcess(submodule, path, command, counter, output_filter, cmd_func).run()


class PFS(object):
    def __init__(self):
        self.meta = Metadata()
        self.__threads = list()
        self.__counter = None

        # Parse arguments provided
        parser = argparse.ArgumentParser()
        parser.add_argument('-v', '--version', action='version', version=self.meta.get_version())
        parser.add_argument('-p', '--path', dest='path', help='System path where execution starts',
                            type=self.exists_path, default=".")
        parser.add_argument('-c', '--command', dest='command', help='Command to execute',
                            # type=self.empty_cmd,
                            default="")
        parser.add_argument('-s', '--schedule', dest='schedule', help='Scheduling strategy', default='load-share',
                            choices=['CHUNK', 'LOAD-SHARE'],
                            type=lambda s: s.upper())
        parser.add_argument('-j', '--jobs', dest='jobs',
                            help='Number of concurrent jobs. Use -j 0 to use automatically the best maximum number of jobs',
                            type=self.valid_jobs, default=2)
        self.__cmd_alias = {
            'pull': ('git pull origin', 'Already up to date', None),
            'status': ('git status', 'nothing to commit', None),
            'pending': ('git log origin/@Abranch@..@Abranch@', '@<empty>@',
                        lambda cmd, tag_value: str(cmd).replace('@Abranch@', tag_value, 2)),
        }
        parser.add_argument('--pull', dest='pull', action='store_true',
                            help='Shortcut to "git pull origin"')
        parser.add_argument('--pending', dest='pending', action='store_true',
                            help='Shortcut to "git log <since origin/current>..<until current>"')
        parser.add_argument('--status', dest='status', action='store_true',
                            help='Shortcut to "git status"')
        parser.add_argument('--verbose', dest='verbose', action='store_true',
                            help='Verbose option in shortcuts')
        self.args = parser.parse_args()

        self.__submodule_path_pattern = re.compile('path ?= ?([A-za-z0-9-_]+)(\/[A-za-z0-9-_]+)*([A-za-z0-9-_])')
        self.__path_pattern = re.compile(' ([A-za-z0-9-_]+)(\/[A-za-z0-9-_]+)*([A-za-z0-9-_])')

    @staticmethod
    def exists_path(path):
        if not os.path.isdir(path):
            raise argparse.ArgumentTypeError("%s is not valid path" % path)
        return path

    @staticmethod
    def empty_cmd(cmd):
        if cmd == "":
            raise argparse.ArgumentTypeError("git cmd is empty")
        return cmd

    @staticmethod
    def valid_jobs(jobs):
        if not str(jobs).isdigit():
            raise argparse.ArgumentTypeError("invalid int value: '%s'" % str(jobs))
        if int(jobs) > multiprocessing.cpu_count() * 2:
            raise argparse.ArgumentTypeError("%s jobs are too many jobs for your computer" % str(jobs))
        if int(jobs) == 0:
            return multiprocessing.cpu_count()
        return int(jobs)

    def read_submodules(self, file):
        submodules = list()
        for line in file:
            submodule_path = re.search(self.__submodule_path_pattern, line)
            if submodule_path:
                path = re.search(self.__path_pattern, submodule_path.group())
                if path:
                    submodules.append(path.group()[1:])  # Remove space at first position
        return submodules

    def run(self):
        start_time = time.time()
        if not os.path.exists(os.path.join(self.args.path, '.gitmodules')):
            print("File .gitmodules not found")
            exit(0)

        with open(os.path.join(self.args.path, '.gitmodules')) as f:
            submodules = self.read_submodules(f)

        self.__counter = Counter(len(submodules))
        print(str(len(submodules)) + " submodules")

        if self.args.schedule == "load-share":
            # Load Share Scheduling
            # ----------------------------
            scheduler = Scheduler(submodules)
        else:
            # Chunk Scheduling
            # ----------------------------
            list_submodule_list = split_load(submodules, self.args.jobs)
            # ----------------------------

        #print(list_submodule_list)

        print("Running with " + str(self.args.jobs) + " threads using " + self.args.schedule + " strategy...")

        command = self.args.command
        output_filter = ""
        command_function = None
        if self.args.pull:
            command = self.__cmd_alias["pull"][0]
            if not self.args.verbose:
                output_filter = self.__cmd_alias["pull"][1]
            command_function = self.__cmd_alias["pull"][2]
        if self.args.status:
            command = self.__cmd_alias["status"][0]
            if not self.args.verbose:
                output_filter = self.__cmd_alias["status"][1]
            command_function = self.__cmd_alias["status"][2]
        if self.args.pending:
            command = self.__cmd_alias["pending"][0]
            if not self.args.verbose:
                output_filter = self.__cmd_alias["pending"][1]
            command_function = self.__cmd_alias["pending"][2]
        try:
            self.empty_cmd(command)
        except argparse.ArgumentTypeError as e:
            print(e)
            exit(0)

        for i in range(self.args.jobs):
            if self.args.schedule == "load-share":
                t = threading.Thread(target=worker,
                                     args=(scheduler, self.args.path, command, self.__counter,
                                           output_filter, command_function,))
            else:
                t = threading.Thread(target=worker,
                                     args=(list_submodule_list[i], self.args.path, command, self.__counter,
                                           output_filter, command_function,))
            self.__threads.append(t)
            t.start()

        for i in range(self.args.jobs):
            self.__threads[i].join()

        print("\nExecution complete!")
        print("--- %s seconds ---" % (time.time() - start_time))


if __name__ == "__main__":
    pfs = PFS()
    pfs.run()



# -*- coding: utf-8 -*-

import subprocess as sub
import os


class PFSProcess(object):
    def __init__(self, submodule, path, cmd, counter, output_filter="", cmd_func=None):
        self.__submodule = submodule
        self.__path = path
        self.__cwd = os.path.join(self.__path, self.__submodule)
        self.__cmd = cmd
        self.__counter = counter
        self.__output_filter = output_filter
        self.__active_branch = self.get_current_branch(self.__cwd)[:-1]
        self.__cmd_func = cmd_func
        self.__output = None
        self.__p = None

    @staticmethod
    def get_current_branch(path):
        return sub.check_output(['git', '-C', path, 'rev-parse', '--abbrev-ref', 'HEAD']).decode('utf-8')

    def run(self):
        is_empty = True
        self.__output = "\n\n" + self.__submodule + "\n"
        # self.__output = sub.check_output(self.__cmd)
        if self.__cmd_func:
            self.__cmd = self.__cmd_func(self.__cmd, self.__active_branch)
        self.__p = sub.Popen(self.__cmd, stdout=sub.PIPE, stderr=sub.PIPE, shell=True,
                             cwd=os.path.join(self.__path, self.__submodule))

        #print(self.__cmd)
        #print(self.__p.communicate()[0].decode('utf-8'))
        if self.__p.communicate()[0].decode('utf-8') != '':
            self.__output += self.__p.communicate()[0].decode('utf-8')  # stdoutdata
            is_empty = False

        if self.__p.communicate()[1]:  # stderrdata
            self.__output += self.__p.communicate()[1].decode('utf-8')

        self.__output += str(self.__counter.increment_value())

        if self.__output_filter == "":
            print(self.__output)
        else:
            if str(self.__output).find(self.__output_filter) == -1 and self.__output_filter != '@<empty>@' \
                    or (str(self.__output).find(self.__output_filter) == -1
                        and self.__output_filter == '@<empty>@' and not is_empty):
                print(self.__output)



# -*- coding: utf-8 -*-

try:
    import queue
except ImportError:
    import Queue as queue   #Python 2 compatibility


class Scheduler(object):
    def __init__(self, submodules):
        self.__queue = queue.Queue()
        self.feed_queue(submodules)

    def feed_queue(self, submodules):

        for submodule in submodules:
            self.__queue.put(submodule)

    def get(self):
        return self.__queue.get()

    def empty(self):
        return self.__queue.empty()



from setuptools import setup
from parallelforeachsubmodule.metadata import Metadata
from io import open

metadata = Metadata()
with open("README.md2rst", "r", encoding="utf-8") as rst_file:
    long_description = rst_file.read()


def requirements():
    """Build the requirements list for this project"""
    requirements_list = []

    with open('requirements.txt') as requirements:
        for install in requirements:
            requirements_list.append(install.strip())

    return requirements_list


setup(
    name='pfs',
    packages=['parallelforeachsubmodule'],
    install_requires=requirements(),
    version=metadata.get_version(),
    license='LGPL v3',
    description='Tool for "git submodule foreach" execution in parallel ',
    long_description=long_description,
    author=metadata.get_author(),
    author_email='contact@rdch106.hol.es',
    url='https://github.com/RDCH106/parallel_foreach_submodule',
    download_url='https://github.com/RDCH106/parallel_foreach_submodule/archive/v'+metadata.get_version()+'.tar.gz',
    entry_points={
        'console_scripts': ['pfs=parallelforeachsubmodule.main:main'],
    },
    keywords='git parallel submodule windows gnu-linux',
    classifiers=['Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7'],
)
